'use strict';

const _ = require('lodash');
const fs = require('fs-extra');
const path = require('path');
const chokidar = require('chokidar');
const EventEmitter = require('events');
const util = require('./util');

const loader = require('./loader');

const TYPES = loader.TYPES,
  CORT_SCHEMA = loadCortSchema(),
  BANK = require('./bank');

module.exports = {
  load,
  loadFile,
  file: loadFile,
  loadFlash,
  flash: loadFlash,
  bank: BANK,
  createDynamicData,
};

function loadCortSchema() {
  var defaultTypes = _.clone(TYPES),
    cortTypes = loadCortTypes(),
    types = defaultTypes.concat(cortTypes),
    schema = loader.Schema.create(
      defaultTypes.concat(cortTypes)
    );
  return schema;
}

function loadCortTypes() {
  var types = [];
  _.each(require('./types'), Pattern => {
    var item = new Pattern(),
      item = Object.assign(item, {
        load: (source, options) => load(source, options),
        loadFile: (filename, options) => loadFile(filename, options),
      }),
      set1 = item.options || {},
      set2 = item.getOptions ? item.getOptions() : {},
      tag = set1.tag || set2.tag || item.tag,

      resolve = set1.resolve || set2.resolve || item.resolve,
      construct = set1.construct || set2.construct || item.construct,
      represent = set1.represent || set2.represent || item.represent,

      options = {
        kind: set1.kind || set2.kind || item.kind,
        instanceOf: set1.instanceOf || set2.instanceOf || item.instanceOf,

        resolve: resolve ? resolve.bind(item) : null,
        construct: construct ? construct.bind(item) : null,
        represent: represent ? represent.bind(item) : null,
      };

    types.push(new loader.Type(tag, options));
  });

  return types;
}

function load(source, options = {}, name) {
  options.schema = options.schema || CORT_SCHEMA;
  var raw = loader.load(source, options),
    data = name ? _.get(raw, name) : raw;
  return data;
}

function loadFile(file, name) {
  var filename = util.pickExt(file),
    source = fs.readFileSync(filename, 'utf-8'),
    data = createDynamicData(
      load(source, { filename }, name),
      filename
    );
  return data;
}

function createDynamicData(data = {}, filename) {

  var handlers = {},
    options = {
      debugMode: false,
      jsonOnUpdate: false,
      jsonFilename: null,
      dumpOnUpdate: false,
      dumpFilename: null,
      filename: filename
    };

  class FlashData {

    constructor (data) {
      var self = this;
      _.extend(self, data);
    }

    getFilename () {
      var self = this;
      return options.filename;
    }

    save () {
      var self = this,
        filename = options.filename;
      if (!filename) { return self; }
      self.dumpToFile(filename);
    }

    debug (val) {
      var self = this;
      options.debugMode = !!val;
    }

    on (name, handler) {
      var self = this;
      handlers[name] = handlers[name] || [];
      handlers[name].push(handler);
    }

    emit (name, ...args) {
      var self = this;
      if (options.debugMode) {
        console.log( 'EMIT>>> ', name, ...args);
      }
      _.each(handlers[name], handler => {
        handler(...args);
      });
    }

    getContext (key) {
      var self = this;
      if (!key) {
        return self;
      }
      return key.includes('.') ? self.get(key.slice(0, key.lastIndexOf('.'))) : self;
    }

    flatten (obj) {
      var self = this,
        item = obj && typeof obj === 'object' ? obj
          : obj && typeof obj === 'string' ? self.get(obj)
            : self,
        data = {},
        flat = util.flatten(item);
      _.each(flat, (val, name) => {
        if (name.includes('__parentFilename') || name.includes('__filename')) {
          return;
        }
        data[name] = val;
      });
      return data;
    }

    set (key, val) {
      var self = this;
      return _.set(self, key, val);
    }

    get (key) {
      var self = this;
      return self.verifyValue(_.get(self, key));
    }

    has (key) {
      if (!key) { return null; }
      var self = this,
        context = self.getContext(key);
      return context && typeof context === 'object'
        && context.hasOwnProperty(key.split('.').pop());
    }

    filename (name) {
      var self = this,
        filename = null,
        p = name;
      while (!filename && p) {
        if (self.has(p) && self.get(p).__filename) {
          return self.get(p).__filename;
        }
        p = p.split('.').slice(0, -1).join('.');
      }
      return filename;
    }

    delete (name) {
      var self = this,
        context = self.getContext(name),
        last = name.split('.').pop(),
        value = _.get(self, name);
      (context && last && context.hasOwnProperty(last)) ? delete context[last] : '';
      self.emit('deleted', name, value);
    }

    change (name, val) {
      var self = this,
        oldValue = _.get(self, name);
      if (oldValue === val) { return; }
      _.set(self, name, val);
      self.emit('changed', name, val, oldValue);
      self.emit('changed:' + name, val, oldValue);
    }

    keys (key) {
      var self = this,
        data = self.get(key);
      return _.keys(self.flatten(data));
    }

    clone () {
      var self = this;
      return _.cloneDeep(self);
    }

    toArray (path, name = 'name') {
      var self = this,
        val = path ? self.get(path) : self;
      if (typeof val !== 'object') {
        return null;
      }
      return util.objectsToArray(val, name);
    }

    clearMeta () {
      var self = this,
        metaKeys = self
          .keys()
          .filter(file => {
            return file.includes('__filename')
              || file.includes('__parentFilename');
          });
      metaKeys.forEach(key => {
        var wrapper = self.get(key.split('.').slice(0, -1).join('.'));
        if (!wrapper) {
          return;
        }
        delete wrapper.__filename;
        delete wrapper.__parentFilename;
      });
      delete self.__filename;
      return self;
    }

    // Handled ===================================================================

    eachPrimitive (handler) {
      var self = this;
      _.each(self.flatten(), (val, name) => {
        handler(val, name);
      });
    }

    each (handler) {
      var self = this;
      return _.each(self, (val, name) => {
        if (name.startsWith('__')) {
          return;
        }
        handler(val, name);
      });
    }

    eachObject (handler, o, parent) {
      var self = this,
        obj = o || self;
      _.keys(obj).forEach(key => {
        var val = obj[key],
          name = parent ? `${parent}.${key}` : key;
        if (util.notObject(val)) {
          return;
        }
        handler(val, name);
        self.eachObject(handler, val, name);
      });
    }

    findKeys (f) {
      var self = this,
        find = f.toLowerCase();
      return self
        .keys()
        .map(key => key.toLowerCase())
        .filter(key => key.indexOf(find) > -1);
    }

    verifyValue (val) {
      var self = this;
      if (!val) { return val; };
      if (util.isObject(val)) {
        return val;
      }
      if (Array.isArray(val)) {
        return _.filter(val);
      }
      return val;
    }

    getParent (key) {
      var self = this,
        parts = key.split('.'),
        jk = parts.pop(),
        p;
      while (parts.length) {
        p = parts.join('.');
        if (self.has(p)) {
          return self.get(p);
        }
        parts.pop();
      }
      return self;
    }

    compileData (keys) {
      var self = this,
        data = {};
      _.each(keys, key => {
        return _.set(data, key, self.get(key));
      });
      return data;
    }

    injectData (data) {
      var self = this,
        keys = self.keys(),
        flatData = self.flatten(data),
        added = [], changed = [], deleted = [], equal = [];
      _.each(flatData, (val, key) => {
        keys.splice(keys.indexOf(key), 1);
        if (!self.has(key)) {
          added.push(key);
          self.set(key, val);
          self.emit('added', key, val);
          self.emit('added:' + key, val);
        } else if (_.isEqual(self.get(key), val)) {
          equal.push(key);
        } else {
          changed.push(key);
          self.set(key, val);
          self.emit('changed', key, val);
          self.emit('changed:' + key, val);
        }
      });
      _.each(keys, key => {
        deleted.push(key);
        self.delete(key);
        self.emit('deleted', key);
        self.emit('deleted:' + key);
      });
      return { added, changed, deleted, equal };
    }

    update (other) {
      var self = this;
      self.injectData(other);
      // if (options.jsonOnUpdate) {
      //   self.jsonToFile(options.jsonFilename);
      // }
      // if (options.dumpOnUpdate) {
      //   self.dumpToFile(options.dumpFilename);
      // }
      self.clearMeta();
      self.emit('updated');
      return self;
    }

    compare (other, keepEqual = false) {
      var self = this,
        f1 = self.flatten(),
        f2 = self.flatten(other),
        deleted = _.difference(_.keys(f1), _.keys(f2)),
        added = _.difference(_.keys(f2), _.keys(f1)),
        changed = _.keys(f2).filter(key => !_.isEqual(f1[key], f2[key]));
      return {
        deleted, added, changed,
        equal: keepEqual ? _.difference(
          _.keys(f1),
          deleted
            .concat(added)
            .concat(changed)
        ) : null
      };
    }

    jsonOnUpdate (filename, force) {
      var self = this;
      options.jsonFilename = filename;
      options.jsonOnUpdate = true;
      force ? self.jsonToFile(filename) : '';
    }

    json () {
      var self = this;
      return JSON.stringify(self, null, 2);
    }

    jsonConsole () {
      var self = this;
      console.log(self.json());
    }

    jsonToFile (filename) {
      var self = this;
      fs.writeFileSync(filename, self.json());
    }

    jsonPipeFiles (config, basedir) {
      var self = this;

      write();
      self.on('updated', write);

      function write() {
        _.each(config, item => {
          var data = {},
            filename = basedir ? path.resolve(basedir, item.file) : item.file;
          _.each(item.props, name => {
            _.extend(data, self.get(name))
          });
          fs.ensureDirSync(path.dirname(filename));
          fs.writeFileSync(
            filename,
            JSON.stringify(data, null, 2)
          );
        });
      }
    }

    dumpOnUpdate (filename) {
      var self = this;
      options.dumpFilename = filename;
      options.dumpOnUpdate = true;
    }

    dump () {
      var self = this;
      return loader.dump(self);
    }

    dumpConsole () {
      var self = this;
      console.log(self.dump());
    }

    dumpToFile (filename) {
      var self = this;
      fs.writeFileSync(filename, self.dump());
    }

    set __filename (val) {
      var self = this;
      options.filename = val;
    }

    set __parentFilename (val) {
      var self = this;
      options.parentFilename = val;
    }

    getParentFilename () {
      var self = this;
      return options.parentFilename;
    }

  }

  var item = new FlashData(data);

  return item;
}

function loadFlash(file, name) {
  var filename = util.pickExt(file),
    data = createDynamicData(),
    files = update().concat([filename]),
    watcher = chokidar.watch(files, {
      persistent: true,
      // ignoreInitial: true,
      // followSymlinks: false,
      // interval: 100,
      // binaryInterval: 100,
    });

  watcher.on('change', () => onChange);
  watcher.on('add', () => onChange);
  watcher.on('unlink', () => onChange);

  function onChange() {
    var newFiles = update();
    watcher
      .unwatch(files)
      .add(newFiles);
  }

  return data;

  function update() {
    BANK.resetRequires();
    data.update(
      loadFile(filename, name)
    );
    return BANK.getRequires();
  }
}
